import os
import typing
import plotnine as gg
import pandas as pd
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import utils
from config import Config
from controller.analysecontroller import AnalyseController, Worker
from gui.ui_mainwindow import Ui_MainWindow


class AnalyseUim(QObject):
    """Class managing the Widget which are common for all Picarro calib sub-tabs."""

    def __init__(
        self,
        ctrl: AnalyseController,
        main_ui: Ui_MainWindow,
        config: Config,
        main_window,
    ):
        super(AnalyseUim, self).__init__()
        self.main_ui = main_ui
        self.main_window = main_window
        self.ctrl = ctrl
        self.config = config

        self.__init_tableview__()

        self.main_ui.analyse_pushbutton_dataset_load.clicked.connect(self.__load_dataset__)
        self.main_ui.analyse_toolbutton_run.clicked.connect(self.__run__)
        self.ctrl.sig_simple_step_finished.connect(self.__display_simple_step__)
        self.ctrl.sig_action_step_finished.connect(self.__display_step_with_button__)

    def __load_dataset__(self):
        # Get the directory containing the dataset's files.
        directory = utils.get_data_path(self.config, "datasets", "dataset")
        if directory == "":
            return

        try:
            self.ctrl.get_dataset_reader(directory)
        except ValueError as e:
            utils.show_popup(str(e), "Could not load data set!")
            return

        self.main_ui.analyse_lineedit_dataset.setText(self.ctrl.current_dataset.dataset_name)

    def __run__(self):
        """Run analyse, in another thread to prevent GUI freezing."""
        self.main_ui.analyse_toolbutton_run.setEnabled(False)
        self.thread = QThread()
        self.worker = Worker(self.ctrl)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.finished.connect(lambda: self.main_ui.analyse_toolbutton_run.setEnabled(True))
        self.thread.start()

    def __init_tableview__(self):
        header = self.main_ui.analyse_tablewidget.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)

    def __display_simple_step__(self, message: str):
        table, row_position = self.__add_row_in_table__()
        table.setItem(row_position, 0, QTableWidgetItem(message))

    def __display_step_with_button__(self, message: str, button_text: str, button_action: str):
        table, row_position = self.__add_row_in_table__()
        table.setItem(row_position, 0, QTableWidgetItem(message))
        button = QPushButton(button_text)
        table.setCellWidget(row_position, 1, button)
        action_func = getattr(self, "__" + button_action + "__")
        button.clicked.connect(action_func)

    def __add_row_in_table__(self) -> typing.Tuple[QTableWidget, int]:
        table = self.main_ui.analyse_tablewidget
        row_position = table.rowCount()
        table.insertRow(row_position)
        table.scrollToBottom()
        return table, row_position

    def __view_collector_df__(self):
        self.__view_dataframe_in_qdialog__(self.ctrl.collector_df)

    def __view_volume_dict__(self):
        volume_dict_df = pd.DataFrame(self.ctrl.volumes_dict, index=["volume_ml"]).T
        self.__view_dataframe_in_qdialog__(volume_dict_df)

    def __view_arrival_df__(self):
        self.__view_dataframe_in_qdialog__(self.ctrl.arrival_df)

    def __view_arrival_df_raw__(self):
        self.__view_dataframe_in_qdialog__(self.ctrl.arrival_df_raw)

    def __open_arrival_csv__(self):
        os.system("xdg-open " + self.ctrl.arrival_filepath)

    def __view_flask_df__(self):
        self.__view_dataframe_in_qdialog__(self.ctrl.flask_df)

    def __open_flask_csv__(self):
        os.system("xdg-open " + self.ctrl.flask_df_filepath)

    def __view_flask_plot__(self):
        flask_df = self.ctrl.flask_df.copy()
        if not all(flask_df["flask"] == 0):
            flask_df["mm_diff"] = flask_df["max"] - flask_df["min"]
            mm_per_flask_df = flask_df[["flask", "mm_diff"]].groupby("flask").agg(sum)
            print(
                gg.ggplot()
                + gg.geom_histogram(
                    data=mm_per_flask_df, mapping=gg.aes(x="mm_diff", bins=round(len(mm_per_flask_df.index) / 3))
                )
            )

    def __open_pic_conduc_height_df__(self):
        os.system("xdg-open " + self.ctrl.pic_conduc_height_filepath)

    def __view_picarro_depth_plot__(self):
        arrival_df = self.ctrl.arrival_df.copy()
        picarro_df = self.ctrl.picarro_df.copy()

        arrival_df = arrival_df[arrival_df["icbk_code"] > 0]
        picarro_df = pd.merge_asof(
            left=arrival_df[["icbk_code", "icbk_name", "melted_height", "picarro"]],
            right=picarro_df,
            left_on="picarro",
            right_index=True,
        )
        picarro_df["iceblock"] = picarro_df["icbk_code"].astype(str) + "-" + picarro_df["icbk_name"]

        print(
            gg.ggplot()
            + gg.geom_path(data=picarro_df, mapping=gg.aes(x="Delta_18_16", y="melted_height", color="iceblock"))
        )

    def __edit_calib_df__(self):
        self.__edit_dataframe_in_qdialog__(self.ctrl.calib_df)

    def __view_calib_df__(self):
        self.__view_dataframe_in_qdialog__(self.ctrl.calib_df)

    def __open_calib_pic_conduc_height_df__(self):
        os.system("xdg-open " + self.ctrl.pic_conduct_calibrated_filepath)

    def __open_calib_flask_df__(self):
        os.system("xdg-open " + self.ctrl.flask_calibrated_filepath)

    def __view_dataframe_in_qdialog__(self, dataframe: pd.DataFrame):
        self.model = TableModel(dataframe)
        self.table = QTableView()
        self.table.setModel(self.model)
        my_dialog = QDialog(self.main_window)
        layout = QVBoxLayout()
        layout.addWidget(self.table)
        my_dialog.setLayout(layout)
        my_dialog.show()

    def __edit_dataframe_in_qdialog__(self, dataframe: pd.DataFrame):
        self.model = EditableTableModel(dataframe)
        self.model.dataChanged.connect(self.__update_model_df__)
        self.table = QTableView()
        self.table.setModel(self.model)
        my_dialog = QDialog(self.main_window)
        layout = QVBoxLayout()
        layout.addWidget(self.table)
        my_dialog.setLayout(layout)
        my_dialog.show()

    def __update_model_df__(self):
        self.ctrl.update_calib_csv(self.model.getData())


class TableModel(QAbstractTableModel):
    def __init__(self, data):
        super(TableModel, self).__init__()
        self._data = data

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            value = self._data.iloc[index.row(), index.column()]
            return str(value)

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, index):
        return self._data.shape[1]

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return str(self._data.columns[section])

            if orientation == Qt.Vertical:
                return str(self._data.index[section])


class EditableTableModel(TableModel):
    def __init__(self, data):
        super(EditableTableModel, self).__init__(data)

    def getData(self):
        return self._data

    def setData(self, index, value, role):
        if not index.isValid():
            return False
        if role != Qt.EditRole:
            return False
        row = index.row()
        if row < 0 or row >= len(self._data.values):
            return False
        column = index.column()
        if column < 0 or column >= self._data.columns.size:
            return False
        self._data.iloc[row, column] = value
        self.dataChanged.emit(index, index)
        return True

    def flags(self, index):
        flags = super(self.__class__, self).flags(index)
        flags |= Qt.ItemIsEditable
        flags |= Qt.ItemIsSelectable
        flags |= Qt.ItemIsEnabled
        flags |= Qt.ItemIsDragEnabled
        flags |= Qt.ItemIsDropEnabled
        return flags
