import glob
import os
import re
from PyQt5.QtCore import *
from PyQt5.QtMultimedia import *
from PyQt5.QtMultimediaWidgets import *
from PyQt5.QtCore import Qt, QUrl

from config import Config
from cfatools.logreader.dataset import DatasetReader

from gui.ui_meltingcam_dialog import Ui_MeltingcamDialog
from uim.icbktransitionuim import IcbktransitionUim
from uim.explouim import ExploUim


class MeltingcamUim(QObject):
    """Class managing the Widget of the MeltingCam Dialog."""

    def __init__(
        self,
        meltingcam_ui: Ui_MeltingcamDialog,
        config: Config,
        explo_uim: ExploUim,
        icbktransition_uim: IcbktransitionUim,
    ):
        super(MeltingcamUim, self).__init__()
        self.meltingcam_ui = meltingcam_ui
        self.config = config

        self.mediaPlayer = QMediaPlayer(None, QMediaPlayer.VideoSurface)
        self.mediaPlayer.positionChanged.connect(self.update_current_time)
        self.mediaPlayer.durationChanged.connect(self.meltingcam_ui.horizontalslider_current.setMaximum)

        # Control buttons
        self.meltingcam_ui.toolbutton_play.clicked.connect(self.mediaPlayer.play)
        self.meltingcam_ui.toolbutton_pause.clicked.connect(self.mediaPlayer.pause)

        # Time slider
        self.meltingcam_ui.horizontalslider_current.valueChanged.connect(self.set_position_from_slider)
        self.meltingcam_ui.timeedit_current.timeChanged.connect(self.set_position_from_timeedit)

        explo_uim.sig_cursor_moved.connect(self.set_position_from_signal)
        icbktransition_uim.sig_cursor_moved.connect(self.set_position_from_signal)
        explo_uim.explo_prvd.sig_dataset_loaded.connect(self.load_video_list)

        self.dataset_path = ""
        self.videoWidget = QVideoWidget()
        self.meltingcam_ui.layout_video.addWidget(self.videoWidget)

        self.meltingcam_ui.combobox_filename.currentTextChanged.connect(self.load_video)

    def load_video_list(self, dataset_reader: DatasetReader):
        self.meltingcam_ui.combobox_filename.clear()
        self.dataset_path = dataset_reader.dataset_path
        video_paths = glob.glob(self.dataset_path + "*_meltingcam.avi")
        video_paths.sort()
        video_filenames = list()
        for video_path in video_paths:
            video_filenames.append(os.path.basename(video_path))
        self.meltingcam_ui.combobox_filename.addItems(video_filenames)

        if len(video_paths) != 0:
            self.load_video(video_filenames[0])

    def load_video(self, video_filename: str):
        time_str = re.search("[0-9]{8}-(.*)_meltingcam.avi", video_filename).group(1)  # Get the "time" part in filename
        self.start_time = QTime.fromString(time_str[:-3], "HHmmss.zzz")  # QTime does not manage microseconds
        self.mediaPlayer.setMedia(QMediaContent(QUrl.fromLocalFile(os.path.join(self.dataset_path, video_filename))))
        self.mediaPlayer.setVideoOutput(self.videoWidget)
        self.mediaPlayer.pause()

    def set_position_from_slider(self, position_msec: int):
        self.mediaPlayer.setPosition(position_msec)
        self.update_current_time(position_msec)

    def set_position_from_timeedit(self, position_time: QTime):
        position_msec = self.start_time.msecsTo(position_time)
        self.mediaPlayer.setPosition(position_msec)
        self.meltingcam_ui.horizontalslider_current.setValue(position_msec)

    def set_position_from_signal(self, position_time: QTime):
        if not self.meltingcam_ui.checkbox_bind.isChecked():
            return
        position_msec = self.start_time.msecsTo(position_time)
        self.mediaPlayer.setPosition(position_msec)
        self.update_current_time(position_msec)
        self.meltingcam_ui.horizontalslider_current.setValue(position_msec)

    def update_current_time(self, position_msec: int):
        current = self.start_time.addMSecs(position_msec)
        self.meltingcam_ui.timeedit_current.setTime(current)
