import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import QColor
import pandas as pd
import numpy as np
from math import atan, degrees

import utils
from uim.picarrocalibuim import PicarroCalibUim


class PicarroCorrectionUim:
    """Class managing the 'Injections identification' sub-tab of the 'Picarro calibration' tab."""

    # "Variables" tableWidget columns identifiers.
    DATETIME_COL = 0
    STANDARD_COL = 1
    H2O_COL = 2
    D18O_COL = 3
    D18OSD_COL = 4
    D2H_COL = 5
    D2HSD_COL = 6
    KEEP_COL = 7

    STD_COLORS = [
        QColor(228, 26, 28),  # Red
        QColor(55, 126, 184),  # Blue
        QColor(77, 175, 74),  # Green
        QColor(152, 78, 163),  # Violet
        QColor(255, 127, 0),  # Orange
        QColor(255, 255, 51),  # Yellow
        QColor(166, 86, 40),  # Brown
        QColor(247, 129, 191),
    ]  # Pink

    def __init__(self, picarrocalib_uim: PicarroCalibUim):
        self.main_ui = picarrocalib_uim.main_ui
        self.picarro_prvd = picarrocalib_uim.picarro_prvd

        self.__init_tablewidget__()
        self.standard_colors = dict()

        self.calibs_df = pd.DataFrame()

        self.main_ui.picarro_spinbox_dayrange_before.valueChanged.connect(
            lambda value: self.picarro_prvd.update_near_days_range("before", value)
        )
        self.main_ui.picarro_spinbox_dayrange_after.valueChanged.connect(
            lambda value: self.picarro_prvd.update_near_days_range("after", value)
        )
        self.main_ui.picarro_pushbutton_correction_save.clicked.connect(
            self.__save_calibration_points_and_equations__
        )

        self.picarro_prvd.sig_near_calibs_df_updated.connect(self.update_table)
        self.picarro_prvd.sig_near_calibs_df_updated.connect(
            self.__update_calibs_timeseries_plot__
        )
        self.picarro_prvd.sig_equation_dict_updated.connect(
            self.__update_equation_plot__
        )
        self.picarro_prvd.sig_equation_dict_updated.connect(
            self.__update_equation_text__
        )

    def __init_colors__(self, calibs_df: pd.DataFrame):
        colors = dict()
        standard_names = calibs_df["standard"].unique()
        for i in range(len(standard_names)):
            colors[standard_names[i]] = self.STD_COLORS[i % len(self.STD_COLORS)]
        self.standard_colors = colors

    ####################################################################
    # Calibrations' table

    def __init_tablewidget__(self):
        table = self.main_ui.picarro_tablewidget_calibrations

        table.verticalHeader().setStyleSheet("QHeaderView { font-size: 10pt; }")
        table.horizontalHeader().setStyleSheet("QHeaderView { font-size: 10pt; }")

        table.verticalHeader().setFixedWidth(30)  # Width of the Row header
        table.setColumnWidth(self.DATETIME_COL, 150)
        table.setColumnWidth(self.STANDARD_COL, 70)
        table.setColumnWidth(self.H2O_COL, 70)
        table.setColumnWidth(self.D18O_COL, 70)
        table.setColumnWidth(self.D18OSD_COL, 60)
        table.setColumnWidth(self.D2H_COL, 70)
        table.setColumnWidth(self.D2HSD_COL, 60)
        table.setColumnWidth(self.KEEP_COL, 50)

        table.horizontalHeader().setFixedHeight(24)  # Height of the Column header

    def update_table(self, table_df: pd.DataFrame):
        # Reset Standards' colors
        self.__init_colors__(table_df)

        # Clear previous table's content
        self.main_ui.picarro_tablewidget_calibrations.clearContents()
        self.main_ui.picarro_tablewidget_calibrations.model().removeRows(
            0, self.main_ui.picarro_tablewidget_calibrations.rowCount()
        )

        # Add rows for the new subset
        for index, row in table_df.iterrows():
            self.__add_row__(index, row)

    def __add_row__(self, row_id: int, row_content: pd.Series):
        table = self.main_ui.picarro_tablewidget_calibrations

        table.insertRow(row_id)

        # Datetime
        datetime_item = QDateTimeEdit(row_content["valid_start"])
        datetime_item.setDisplayFormat("dd/MM/yyyy HH:mm")
        datetime_item.setReadOnly(True)
        table.setCellWidget(row_id, self.DATETIME_COL, datetime_item)

        # Standard name
        standard_item = QTableWidgetItem()
        standard_item.setFlags(Qt.ItemIsEnabled)  # Read only
        standard_item.setBackground(self.standard_colors[row_content["standard"]])
        standard_item.setText(row_content["standard"])
        table.setItem(row_id, self.STANDARD_COL, standard_item)

        # Injections' values
        table.setItem(
            row_id, self.H2O_COL, self.__get_number_item__(row_content["H2O_mean"])
        )
        table.setItem(
            row_id,
            self.D18O_COL,
            self.__get_number_item__(row_content["Delta_18_16_mean"]),
        )
        table.setItem(
            row_id,
            self.D18OSD_COL,
            self.__get_number_item__(row_content["Delta_18_16_std"]),
        )
        table.setItem(
            row_id,
            self.D2H_COL,
            self.__get_number_item__(row_content["Delta_D_H_mean"]),
        )
        table.setItem(
            row_id,
            self.D2HSD_COL,
            self.__get_number_item__(row_content["Delta_D_H_std"]),
        )

        # Keep
        keep_item = QCheckBox()
        keep_item.setChecked(row_content["keep"])
        table.setCellWidget(row_id, self.KEEP_COL, keep_item)
        table.cellWidget(row_id, self.KEEP_COL).stateChanged.connect(
            lambda state, row_id=row_id: self.picarro_prvd.mark_calib(
                row_id, bool(state)
            )
        )

    def __get_number_item__(self, value: float) -> QTableWidgetItem:
        number_item = QTableWidgetItem()
        number_item.setFlags(Qt.ItemIsEnabled)  # Read only
        number_item.setText("{:.2f}".format(value))
        return number_item

    ####################################################################
    # Calibrations' timeseries plot

    def __update_calibs_timeseries_plot__(self):
        self.__init_colors__(self.picarro_prvd.near_calibs_df)

        # Create a layout
        layout = pg.GraphicsLayout()
        layout.layout.setSpacing(0.0)
        layout.setContentsMargins(0.0, 0.0, 0.0, 0.0)

        # Add it to the graphics view
        self.main_ui.picarro_graphicsview_time.setCentralItem(layout)

        # Create plot items and add them to the layout
        plot_delta_18_16 = self.__get_timeseries_plot_item__(layout, 0, "Delta_18_16")
        plot_delta_D_H = self.__get_timeseries_plot_item__(layout, 1, "Delta_D_H")
        plot_delta_D_H.setXLink(plot_delta_18_16)

    def __get_timeseries_plot_item__(
        self, layout: pg.GraphicsLayout, rank: int, species: str
    ):
        calibs_df = self.picarro_prvd.near_calibs_df
        plot_item = layout.addPlot(
            rank, 0, axisItems={"bottom": utils.TimeAxisItem(orientation="bottom")}
        )

        # Add runs points
        for standard in calibs_df["standard"].unique():
            standard_df = calibs_df[
                (calibs_df["standard"] == standard) & (calibs_df["keep"])
            ]
            centered_mean = (
                standard_df[species + "_mean"] - standard_df[species + "_mean"].mean()
            )
            run_points = pg.ScatterPlotItem(
                x=utils.pd_time_to_epoch_ms(standard_df["valid_start"]),
                y=centered_mean,
                size=6,
                brush=self.standard_colors[standard],
                name=standard,
            )
            plot_item.addItem(run_points)

            err = pg.ErrorBarItem(
                x=np.array(utils.pd_time_to_epoch_ms(standard_df["valid_start"])),
                y=np.array(centered_mean),
                height=np.array(standard_df[species + "_std"]) * 2,
                beam=60 * 5,  # 5 minutes
                pen=self.standard_colors[standard],
            )
            plot_item.addItem(err)

        plot_item.setLabel("left", species)
        return plot_item

    ####################################################################
    # Calibrations' correction line plot

    def __update_equation_plot__(self):
        self.__init_colors__(self.picarro_prvd.near_calibs_df)

        # Create a layout
        layout = pg.GraphicsLayout()
        layout.layout.setSpacing(0.0)
        layout.setContentsMargins(0.0, 0.0, 0.0, 0.0)

        # Add it to the graphics view
        self.main_ui.picarro_graphicsview_equation.setCentralItem(layout)

        # Create plot items and add them to the layout
        plot_delta_18_16 = self.__get_equation_plot_item__(layout, 0, "Delta_18_16")
        plot_delta_D_H = self.__get_equation_plot_item__(layout, 1, "Delta_D_H")

    def __get_equation_plot_item__(
        self, layout: pg.GraphicsLayout, rank: int, species: str
    ):
        plot_item = layout.addPlot(rank, 0)

        # Add runs points
        for standard in self.picarro_prvd.near_calibs_df["standard"].unique():

            standard_df = self.picarro_prvd.near_calibs_df[
                (self.picarro_prvd.near_calibs_df["standard"] == standard)
                & (self.picarro_prvd.near_calibs_df["keep"])
            ]

            # Get the standard's true value
            stdtrue_mean = standard_df[species + "_truevalue"].to_list()
            stdtrue_std = standard_df[species + "_truevalue_std"].to_list()

            run_points = pg.ScatterPlotItem(
                x=standard_df[species + "_mean"],
                y=stdtrue_mean,
                size=6,
                brush=self.standard_colors[standard],
                name=standard,
            )
            plot_item.addItem(run_points)

            err = pg.ErrorBarItem(
                x=np.array(standard_df[species + "_mean"]),
                y=np.array(stdtrue_mean),
                width=np.array(stdtrue_std) * 2,
                height=np.array(standard_df[species + "_std"]) * 2,
                pen=self.standard_colors[standard],
            )
            plot_item.addItem(err)

        equation_dict = self.picarro_prvd.equation_dict[species]
        reg_line = pg.InfiniteLine(
            pos=QPointF(0, equation_dict["intercept"]),
            angle=degrees(atan(equation_dict["slope"])),
        )
        plot_item.addItem(reg_line)

        plot_item.setLabel("left", species)
        return plot_item

    def __update_equation_text__(self):
        for species in ["Delta_18_16", "Delta_D_H"]:
            lineedit = getattr(self.main_ui, "picarro_lineedit_equation_" + species)
            equation_dict = self.picarro_prvd.equation_dict[species]
            text = (
                "True = "
                + "{:.5f}".format(equation_dict["slope"])
                + " * measure "
                + "{:+.5f}".format(equation_dict["intercept"])
            )
            lineedit.setText(text)

    def __save_calibration_points_and_equations__(self):
        filepaths_dict = self.picarro_prvd.save_calibration_points_and_equations()

        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setWindowTitle("Calibration files created")
        msg.setText(
            "The calibration files have successfully been created in:\n - "
            + filepaths_dict["calibs"]
            + "\n - "
            + filepaths_dict["equations"]
        )
        msg.exec_()
